;*** TWI.inc ***;
;------------------------------------------------
; Author: Torsten Knorr
;------------------------------------------------

 .ifndef __RFM12BP_TWI__
 .equ    __RFM12BP_TWI__ = 1

 .include "RFM12BP.inc"

;------------------------------------------------

 .equ TWI_SLA_W_ACK      = 0x60
 .equ TWI_SLA_W_DATA_ACK = 0x80
 .equ TWI_SLA_W_STOP     = 0xA0

 .equ TWI_SLA_R_ACK      = 0xA8
 .equ TWI_SLA_R_DATA_ACK = 0xB8
 .equ TWI_SLA_R_NACK     = 0xC0

;------------------------------------------------

 .macro M_INIT_TWI

; clear TWI bit rate register
    clr  R_TEMP_L
    sts  TWBR,          R_TEMP_L

; clear TWI slave address mask register
    sts  TWAMR,         R_TEMP_L

; clear TWI prescaler bits
    sts  TWSR,          R_TEMP_L

; set TWI slave address register
    ldi  R_TEMP_L,      TWI_SLAVE_ADDR
    sts  TWAR,          R_TEMP_L

    ldi  R_TEMP_L,      ((1 << TWEN) | (1 << TWEA) | (1 << TWIE) | (1 << TWINT))
    sts  TWCR,          R_TEMP_L

    cbi  STATUS_REG,    TWI_BUSY_BIT
    cbi  STATUS_REG,    TWI_REQUEST

 .endmacro

;------------------------------------------------

 .macro TWI_ENABLE

    ldi  R_TEMP_L,      ((1 << TWEN) | (1 << TWEA) | (1 << TWIE) | (1 << TWINT))
    sts  TWCR,          R_TEMP_L

    cbi  STATUS_REG,    TWI_BUSY_BIT
    cbi  STATUS_REG,    TWI_REQUEST

 .endmacro

;------------------------------------------------

 .macro TWI_WAIT_BUSY

 TWI_WAIT_BUSY_LOOP:
    sbic STATUS_REG,    TWI_BUSY_BIT
    rjmp TWI_WAIT_BUSY_LOOP

 .endmacro

;------------------------------------------------

 .macro TWI_DISABLE

    TWI_WAIT_BUSY
    
    clr  R_TEMP_L
    sts  TWCR,          R_TEMP_L

 .endmacro

;------------------------------------------------

 .macro TWI_DISABLE_ACK

    TWI_WAIT_BUSY

    lds  R_TEMP_L,      TWCR
    cbr  R_TEMP_L,      (1 << TWEA)
    sts  TWCR,          R_TEMP_L

 .endmacro

;------------------------------------------------

 .macro TWI_ENABLE_ACK

    lds  R_TEMP_L,      TWCR
    sbr  R_TEMP_L,      (1 << TWEA)
    sts  TWCR,          R_TEMP_L

 .endmacro

;------------------------------------------------

 .ifdef __RFM12BP_RESPONSE__

 .macro M_RFM12BP_TWI_INT_RESPONSE

    in   RI_SREG,       SREG

    lds  RI_TWI,        TWSR

; Own SLA+W has been received ACK has been returned
    cpi  RI_TWI,        TWI_SLA_W_ACK
    breq TWI_STATUS_SLA_W_ACK

; Previously addressed with own SLA+W data has been received ACK has been returned
    cpi  RI_TWI,        TWI_SLA_W_DATA_ACK
    breq TWI_STATUS_SLA_W_DATA_ACK

; A STOP condition or repeated START condition has been received while still addressed as slave
    cpi  RI_TWI,        TWI_SLA_W_STOP
    breq TWI_STATUS_SLA_W_STOP

    rjmp TWI_STATUS_OK

;------------------------------------------------
; Own SLA+W has been received ACK has been returned
 TWI_STATUS_SLA_W_ACK:

    ldi  RI_YP_TWI_L,   low(s_finale)
    ldi  RI_YP_TWI_H,   high(s_finale)
    movw RI_YP_END_L,   RI_YP_TWI_L

    ldi  RI_YP_TWI_L,   low(s_data)
    ldi  RI_YP_TWI_H,   high(s_data)
    sbi  STATUS_REG,    TWI_BUSY_BIT
    rjmp TWI_STATUS_OK

;------------------------------------------------
; Previously addressed with own SLA+W data has been received ACK has been returned
 TWI_STATUS_SLA_W_DATA_ACK:

; check pointer     >=
    cp   RI_YP_TWI_L,   RI_YP_END_L
    cpc  RI_YP_TWI_H,   RI_YP_END_H
    brsh TWI_STATUS_OK

    lds  RI_TWI,        TWDR
    st   Y+,            RI_TWI

    rjmp TWI_STATUS_OK

;------------------------------------------------
; A STOP condition or repeated START condition has been received while still addressed as slave
 TWI_STATUS_SLA_W_STOP:
    cbi  STATUS_REG,    TWI_BUSY_BIT

 TWI_STATUS_OK:
    lds  RI_TWI,        TWCR
    sbr  RI_TWI,        (1 << TWINT)
    sts  TWCR,          RI_TWI

    out  SREG,          RI_SREG
    reti

 .endmacro

 .endif

;------------------------------------------------

 .ifdef __RFM12BP_REQUEST__

 .macro M_RFM12BP_TWI_INT_REQUEST

    in   RI_SREG,       SREG

    lds  RI_TWI,        TWSR

; Own SLA+W has been received ACK has been returned
    cpi  RI_TWI,        TWI_SLA_W_ACK
    breq TWI_STATUS_SLA_W_ACK

; Previously addressed with own SLA+W data has been received ACK has been returned
    cpi  RI_TWI,        TWI_SLA_W_DATA_ACK
    breq TWI_STATUS_SLA_W_DATA_ACK

; A STOP condition or repeated START condition has been received while still addressed as slave
    cpi  RI_TWI,        TWI_SLA_W_STOP
    breq TWI_STATUS_SLA_W_STOP

; Own SLA+R has been received ACK has been returned
    cpi  RI_TWI,        TWI_SLA_R_ACK
    breq TWI_STATUS_SLA_R_ACK

; Data byte in TWDR has been transmitted; ACK has been received
    cpi  RI_TWI,        TWI_SLA_R_DATA_ACK
    breq TWI_STATUS_SLA_R_DATA_ACK

; Data byte in TWDR has been transmitted; NOT ACK has been received
    cpi  RI_TWI,        TWI_SLA_R_NACK
    breq TWI_STATUS_SLA_R_NACK

    rjmp TWI_STATUS_OK

;------------------------------------------------
; Own SLA+W has been received ACK has been returned
 TWI_STATUS_SLA_W_ACK:

    ldi  RI_YP_TWI_L,   low(s_finale)
    ldi  RI_YP_TWI_H,   high(s_finale)
    movw RI_YP_END_L,   RI_YP_TWI_L

    ldi  RI_YP_TWI_L,   low(s_identifier)
    ldi  RI_YP_TWI_H,   high(s_identifier)
    sbi  STATUS_REG,    TWI_BUSY_BIT
    rjmp TWI_STATUS_OK

;------------------------------------------------
; Previously addressed with own SLA+W data has been received ACK has been returned
 TWI_STATUS_SLA_W_DATA_ACK:

; check pointer     >=
    cp   RI_YP_TWI_L,   RI_YP_END_L
    cpc  RI_YP_TWI_H,   RI_YP_END_H
    brsh TWI_STATUS_OK

    lds  RI_TWI,        TWDR
    st   Y+,            RI_TWI

    rjmp TWI_STATUS_OK

;------------------------------------------------
; A STOP condition or repeated START condition has been received while still addressed as slave
 TWI_STATUS_SLA_W_STOP:

    cbi  STATUS_REG,    TWI_BUSY_BIT
    sbi  STATUS_REG,    TWI_REQUEST

    rjmp TWI_STATUS_OK

;------------------------------------------------
; Own SLA+R has been received ACK has been returned
 TWI_STATUS_SLA_R_ACK:

    ldi  RI_YP_TWI_L,   low(s_rx_buff_end)
    ldi  RI_YP_TWI_H,   high(s_rx_buff_end)
    movw RI_YP_END_L,   RI_YP_TWI_L

    ldi  RI_YP_TWI_L,   low(s_rx_buff)
    ldi  RI_YP_TWI_H,   high(s_rx_buff)
    sbi  STATUS_REG,    TWI_BUSY_BIT

    ld   RI_TWI,        Y+
    sts  TWDR,          RI_TWI
    rjmp TWI_STATUS_OK

;------------------------------------------------
; Data byte in TWDR has been transmitted; ACK has been received
 TWI_STATUS_SLA_R_DATA_ACK:

; check pointer     >=
    cp   RI_YP_TWI_L,   RI_YP_END_L
    cpc  RI_YP_TWI_H,   RI_YP_END_H
    brsh TWI_STATUS_OK

    ld   RI_TWI,        Y+
    sts  TWDR,          RI_TWI
    rjmp TWI_STATUS_OK

;------------------------------------------------
; Data byte in TWDR has been transmitted; NOT ACK has been received
 TWI_STATUS_SLA_R_NACK:

    cbi  STATUS_REG,     TWI_BUSY_BIT

    ;rjmp TWI_STATUS_OK

;------------------------------------------------

 TWI_STATUS_OK:

    lds  RI_TWI,        TWCR
    sbr  RI_TWI,        (1 << TWINT)
    sts  TWCR,          RI_TWI

    out  SREG,          RI_SREG
    reti

 .endmacro

 .endif

;------------------------------------------------

 .endif

;------------------------------------------------
